/* sftest.c
 * ============================================================
 * Demo-Programm f|r formatierte Eingabe (95/04/27)
 * Copyright 1995: Martin Weitzel, Darmstadt
 * ============================================================
*/

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CLIMIT 200

int cget(	/* -- Einlesen + Ausgeben con char[] */
	const char *ifmt,	/* Einleseformat */
	const char *ofmt	/* Ausgabeformat */
) {
	char v[CLIMIT]; int r;
	(void) memset(v, '\0', CLIMIT);
	r = scanf(ifmt, v);
	(void) printf("(%d)", r);
	if (r > 0)
		(void) printf(ofmt, v);
	return r;
}
/*
 * Hilfsmakro zur Definition von sechs weiteren,
 * sehr aehnlichen Einlesefunktionen ...
*/
#define	DEFGET(name, type)\
	int name(\
		const char *ifmt,\
		const char *ofmt\
	) {\
		type v; int r;\
		r = scanf(ifmt, &v);\
		(void) printf("(%d)", r);\
		if (r > 0)\
			(void) printf(ofmt, v);\
		return r;\
	}
/*
 * ... und Definition dieser Funktionen:
*/
	DEFGET(	sget,	short	)
	DEFGET(	iget,	int	)
	DEFGET(	lget,	long	)
	DEFGET(	fget,	float	)
	DEFGET(	dget,	double	)
#ifndef NO_LONGDOUBLE
	DEFGET(	ldget,	long double	)
#endif
#undef	DEFGET

typedef struct {	/* Formatspezifikationen */
	const char *convspec;
	int (*gfunc)(const char *, const char *);
	const char *prformat;
} FDESC;

const char GFMT[] = "Lg";	/* Gleitpunktausgabeformat */
const FDESC tab[] = {		/* Formatspezifikationen */
	{  "c",	cget,	"s"	},
	{  "s",	cget		},
	{  "]",	cget,	"s"	},

	{  "d",	iget		},
	{  "i",	iget		},
	{  "o",	iget		},
	{  "u",	iget		},
	{  "x",	iget		},
	{  "X",	iget		},

	{ "hd",	sget		},
	{ "hi",	sget		},
	{ "ho",	sget		},
	{ "hu",	sget		},
	{ "hx",	sget		},
	{ "hX",	sget		},

	{ "ld",	lget		},
	{ "li",	lget		},
	{ "lo",	lget		},
	{ "lu",	iget		},
	{ "lx",	iget		},
	{ "lX",	iget		},

	{  "e",	fget,	GFMT+1	},
	{  "E",	fget,	GFMT+1	},
	{  "f",	fget,	GFMT+1	},
	{  "g",	fget,	GFMT+1	},
	{  "G",	fget,	GFMT+1	},

	{ "le",	dget,	GFMT+1	},
	{ "lE",	dget,	GFMT+1	},
	{ "lf",	dget,	GFMT+1	},
	{ "lg",	dget,	GFMT+1	},
	{ "lG",	dget,	GFMT+1	},
#ifndef NO_LONGDOUBLE
	{ "Le",	ldget,	GFMT	},
	{ "LE",	ldget,	GFMT	},
	{ "Lf",	ldget,	GFMT	},
	{ "Lg",	ldget,	GFMT	},
	{ "LG",	ldget,	GFMT	},
#endif
	0 /* <--- NEVER DELETE THIS!!! */
};

const char *getcvend(	/* -- Ende der Umwandlungsspez. */
	const char *f		/* Zeiger in Formatstring */
) {
	if (*f == '[') {
		if (*++f == '^') ++f;
		if (*f == ']') ++f;
		f = strchr(f, ']');
	}
	return f;
}

const FDESC *fdescfind(	/* -- Tabelleneintrag auswaehlen */
	const char *cs		/* Udwandlungsspezifikation */
) {
	const FDESC *p;
	for (p = tab; p->convspec; ++p) {
		if (!strncmp(cs, p->convspec, strlen(p->convspec)))
			return p;
	}
	return (FDESC *)0;
}

const char *PROG;	/* Programmname */

/*PRINTFLIKE1*/
void failure(		/* -- Abbruch bei fatalen Fehlern */
	const char *message,	/* Meldungstext */
	...			/* Meldungsparameter */
) {
	va_list ap;
	(void) fprintf(stderr, "%s: ", PROG);
	va_start(ap, message);
	(void) vfprintf(stderr, message, ap);
	va_end(ap);
	(void) fprintf(stderr, "\n");
	exit(EXIT_FAILURE);
	/*NOTREACHED*/
}

int moredata(		/* -- Eingabedaten lesen */
	const char *iformat,	/* Formatstring */
	const FDESC *fdesc	/* Formatspezifikation */
) {
	char oformat[5];
	(void) sprintf(oformat, " %%%.2s", fdesc->prformat
					? fdesc->prformat
					: fdesc->convspec);
	return (*fdesc->gfunc)(iformat, oformat) != EOF;
}

int main(		/* -- Hauptprogramm */
	int argc,		/* Argumentzahl */
	char *argv[]		/* Argumentvektor */
) {
	const FDESC *fdesc = (const FDESC *)0;
	const char *f;

	PROG = argv[0];

	if (argc != 2) {
		failure("1 argument required");
		/*NOTREACHED*/
	}
	for (f = argv[1]; (f = strchr(f, '%')) != (char *)0;) {
		if (*++f == '%')
			++f;
		else {
			int no_assign = 0;
			const FDESC *p;
			if (*f == '*') {
				no_assign = 1;
				++f;
			}
			while (isdigit(*f))
				++f;
			if ((f = getcvend(f))
					== (const char *)0) {
				failure("unterminated scanset");
				/*NOTREACHED*/
			}
			if ((p = fdescfind(f))
					== (const FDESC *)0) {
				failure("unsupported conversion");
				/*NOTREACHED*/
			}
			if (no_assign)
				continue;
			if (fdesc != (const FDESC *)0) {
				failure("multiple conversions");
				/*NOTREACHED*/
			}
			fdesc = p;
		}
	}

	while (moredata(argv[1], fdesc)) {
		int c;
		(void) printf(" |");
		while ((c = getchar()) != EOF) {
			if (c != '\n')
				(void) printf("%c", c);
			else {
				(void) printf("\\n\n");
				break;
			}
		}
	}
	(void) printf("\n");

	return EXIT_SUCCESS;
}
